#All R code provided is the intellectual property of Caloo Ecological Services. It is to 
#be used solely for the stated purpose (Adrian Tait's email of 01/10/2014 to Andrea
#Taylor, Xodus) of allowing MSS, SNH and JNCC to check the calculations of 
#collision risk mortality against the standard calculations in the downloadable excel 
#spreadsheets. It is only to be used in the context of determining consent with respect
#to the Hywind Scotland Pilot Park project and is not to be shared with organisations
#outside MS-LOT, MSS, SNH and JNCC.
##############################################################################################################################################
#                                                                                                                                            #     
# This file holds the R functions used in calculating the collision risk mortality estimates.                                                #
#                                                                                                                                            #
##############################################################################################################################################
#
#This is the top level function which provides monthly and annual collision risk estimates for a given scenario
#
run.monthly.SOSS.collision.risk.model<-function(parameters.file,scenario)                                                                    
{

  parameters<-read.csv(parameters.file,stringsAsFactors=FALSE)
  parameters<-parameters[parameters$scenario==scenario,]
  monthly.parameters<-read.csv(parameters$monthly.parameters.file,stringsAsFactors=FALSE)
  blade.profile<-read.csv(parameters$blade.profile.file,stringsAsFactors=FALSE)

  ##############################################################################
  # Calculate total time birds active during operation of turbines for each month
  site.latitude<-parameters$site.latitude
  day.light.definition<-parameters$day.light.definition
  months<-monthly.parameters
  
  yday<-1:365
  date<-as.POSIXct(strptime("01/01/2013",format="%d/%m/%Y")) +(yday-1)*24*60*60
  month<-as.POSIXlt(date)$mon+1
  day.length<-get.day.length(yday,site.latitude,day.light.definition)
  night.length<-24-day.length
  day.lengths<-data.frame(yday,date,month,day.length,night.length)
  months$total.daytime<-tapply(day.lengths$day.length,list(day.lengths$month),sum)
  months$nday<-tapply(day.lengths$day.length,list(day.lengths$month),length)
  months$total.nighttime<-24*months$nday-months$total.daytime
  months$active.nighttime<-months$p.night.active*months$total.nighttime
  months$total.active.time<-months$total.daytime+months$active.nighttime
  months$total.active.time.during.operation<-months$total.active.time*months$Qop
  months$p.mortality.nocturnal<-months$active.nighttime/months$total.active.time
  months$T<-months$total.active.time.during.operation*60*60   # express total time birds active during operation in seconds
  ###################################################################################
  
  months$Da<-months$density.estimate/10^6   # express density as nos/m sq.
  
  # The band options used in the Band 2012 guidance, and the accompanying spreadsheet.
  # band.option==1. Basic, Q2r fixed value
  # band.option==2. Basic, Q2r estimated from flight height distribution
  # band.option==3. Extended model
  band.option<-parameters$band.option
  band<-as.logical(parameters$band)   # IF band==TRUE, provides stage 1 estimates, if band==FALSE, provides stage 2 estimates
  N<-parameters$n.turbine
  R<-parameters$turbine.radius
  hub.height<-parameters$hub.height
  n.blades<-parameters$n.blades
  wingspan<-parameters$wingspan
  body.length<-parameters$body.length
  v<-parameters$flight.speed.m.sec
  flapping<-as.logical(parameters$flapping)
  avoidance.rate<-eval(parse(text=parameters$avoidance.rate))
  xinc<-parameters$xinc
  yinc<-parameters$yinc
  soss.file<-parameters$soss.file
  soss.var<-parameters$soss.var
  max.chord.width<-parameters$max.chord.width
  pitch<-parameters$pitch
  blade.profile$r<-blade.profile$pR*R
  blade.profile$c<-blade.profile$pC*max.chord.width
  blade.profile$pitch<-rep(pitch,length(blade.profile$r))

 T<-months$T
 d<-months$Da
 rpm<-months$rpm
 n.months<-length(months[,1])
 p.collide<-rep(0,n.months) 
 n.flight<-2*T*N*R*v*d
 band.flux.factor<-n.flight*pi/(4*months$Qop)
 turbine.bottom<-hub.height-R
 turbine.top<-hub.height+R
 if(band.option==1)
 {
  Q2r<-parameters$Q2r.fixed
 }else
 {
   fhd.distribution<-"SOSS"
   best.fit<-get.soss.flight.height.distribution(soss.file,soss.var)
   if(band)
   {
     # Take height as centre of each flight height band
     best.fit$height<-best.fit$height-0.5 
     best.fit$cdf<-NULL
     Q2r<-calculate.Q2r.band(R,hub.height,yinc,best.fit)
   }
   else
   {
      Q2r<-calculateQ2r(turbine.bottom,turbine.top,fhd.distribution,best.fit)
   } 

 }
 if(band.option==3)
 {
  extended.collision.risk.model<-TRUE
 }
 else
 {
   extended.collision.risk.model<-FALSE
 }
 
 if(extended.collision.risk.model)
 {
     if(band)
    {
       flux.integral<-calculate.flux.integral(R,hub.height,yinc,best.fit)
      p.of.birds.through.rotor<-flux.integral*pi/4
      for(i in 1:n.months)
      {
        collision.integration<-calculate.collision.integral(R,blade.profile,n.blades,body.length,wingspan,flapping,v,rpm[i],best.fit,hub.height,xinc,yinc)
        collision.integral<-collision.integration$collision.integral
        p.collide[i]<-collision.integral*(pi/4)/p.of.birds.through.rotor
      }  
    } else
    {
      y<-calculate.y(yinc)
      py<-calculate.py(y,hub.height,R,fhd.distribution=fhd.distribution,best.fit)
      p.of.birds.through.rotor<-sum(py)
      for(i in 1:n.months)
      {
        p.collide[i]<-calculate.p.collision.extended(R,blade.profile,n.blades,body.length,wingspan,flapping,v,rpm[i],hub.height,xinc,y,py)
      }
    }  
     p.of.vulnerable.birds.through.rotor<-p.of.birds.through.rotor/Q2r
     
     #
     # CALCULATE flux integral and collision integral retrospectively for stage 2 software purely to allow comparision with stage 1 software
     #
     flux.integral<-p.of.birds.through.rotor*4/pi
     collision.integral<-p.collide[i]*p.of.birds.through.rotor*4/pi
  }   
  else{
      for(i in 1:n.months)
      {
        p.collide[i]<-calculate.p.collision(R,blade.profile,n.blades,body.length,wingspan,flapping,v,rpm[i],n=20)$p.overall
      }  
    p.of.vulnerable.birds.through.rotor<-pi/4
    p.of.birds.through.rotor<-Q2r*p.of.vulnerable.birds.through.rotor
}
  n.vulnerable.birds<-n.flight*Q2r
  n.transit<-n.vulnerable.birds*p.of.vulnerable.birds.through.rotor
  n.collision.no.avoidance<-n.transit*p.collide
  
  months$n.flight<-n.flight  
  months$band.flux.factor<-band.flux.factor
  months$n.vulnerable.birds<-n.vulnerable.birds
  months$n.transit<-n.transit
  months$p.collide<-p.collide
  months$n.collision.no.avoidance<-n.collision.no.avoidance

  avoidance.rate.names<-paste("n.collision.avoidance",avoidance.rate,sep="")
  n.avoidance.rates<-length(avoidance.rate)
  for(i in 1:n.avoidance.rates)
  {
    months[,avoidance.rate.names[i]]<-n.collision.no.avoidance*(1-avoidance.rate[i])
  }
  overall<-parameters
  overall$Q2r<-Q2r
  overall$p.of.vulnerable.birds.through.rotor<-p.of.vulnerable.birds.through.rotor
  overall$p.of.birds.through.rotor<-p.of.birds.through.rotor
  overall$band.flux.factor<-sum(months$band.flux.factor)
  overall$n.flight<-sum(months$n.flight)
  overall$n.vulnerable.birds<-sum(months$n.vulnerable.birds)
  overall$n.transit<-sum(months$n.transit)
  overall$n.collision.no.avoidance<-sum(months$n.collision.no.avoidance)
  overall$average.p.collide<-overall$n.collision.no.avoidance/overall$n.transit
  
  overall$p.mortality.nocturnal<-sum(months$p.mortality.nocturnal*months$n.collision.no.avoidance)/overall$n.collision.no.avoidance
  overall$flux.integral<-p.of.birds.through.rotor*4/pi
  overall$collision.integral<-overall$average.p.collide*p.of.birds.through.rotor*4/pi

  for(i in 1:n.avoidance.rates)
  {
    overall[,avoidance.rate.names[i]]<-sum(months[,avoidance.rate.names[i]])
  }
 list(months=months,overall=overall)
}
#
# This function retrieves a SOSS generic flight height distribution from file, and prepares it for analysis.
# This includes calculating the cumulative distribution function, which is used to provide a common theoretical basis for collisiom risk modelling across
# all types of flight height distribution
#
get.soss.flight.height.distribution<-function(soss.file,soss.var)      
{         
   all.fhd<-read.csv(soss.file,stringsAsFactors=FALSE)
   p<-all.fhd[,soss.var]
   height<-all.fhd$V2
   cdf<-cumsum(p)                                                      
   fhd<-data.frame(height=c(0,height),p=c(0,p),cdf=c(0,cdf))
   fhd
}
#
# This function calculates the day length for a given day of the year at a given latitude by implementing the equation given in Forsythe et al. 1995
# By default it uses day light definition 3, the definition used in the Band spreadsheet, which excludes twilight from daylight hours.
# For the Hywind project we have used daylight definition 4, which includes civic twilight within daylight hours
#
get.day.length<-function(yday,latitude,day.light.definition=3)
{
  ps<-c(0.0,2.6667,0.8333,6.0,12.0,18.0)
  p<-ps[day.light.definition]
  latitude<-latitude*pi/180
  solar.declination<-asin(0.39795*cos(0.2163108+ 2*atan(0.9671396*tan(0.00860*(yday-186)))))
  day.length<- 24 -(24/pi)*acos((sin(p*pi/180)+sin(latitude)*sin(solar.declination))/(cos(latitude)*cos(solar.declination)))
  day.length
}  
#
#  This function calculates py, the proportion of all flying birds passing through the rotor within each height band. Used by the stage 2 software for the extended model
#
calculate.py<-function(y,hub.height,R,fhd.distribution,best.fit)
{
 x<-sqrt(1-y^2)
 triangle.area<-x*abs(y)
 segment.area<-acos(abs(y))
 area.outside.chord<-segment.area-triangle.area
 area.below.chord<-pi-area.outside.chord
 area.below.chord[y<0]<-area.outside.chord[y<0]
 # p1 is the proportion of birds in each height band heading through the rotor
 p1<-diff(area.below.chord)/(2*diff(y))
 height<-y*R+hub.height
 # p2 is the proportion of birds in each height band
 p2<-diff(get.cdf(x=height,distribution=fhd.distribution,best.fit=best.fit))
 #py is the proportion of all flying birds passing through the rotor at a given height
 py<-p1*p2
 py
}
#
# This function calculates y coordinates for the boundaries of the flight height bands used in the extended mode for both the stage 1 and stage 2 software
#
calculate.y<-function(yinc)
{
  ny <-ceiling(1/yinc)    
  yinc<-1 /ny
  seq(-1,1,yinc)
}
#
# This functiom executes numerical integration using the trapezium rule
#
trap.rule <- function(x,y)
{
  sum(diff(x)*(y[-1]+y[-length(y)]))/2 
} 
#
# This function calculates the probability of collision for the extended model in the stage 2 software
#
calculate.p.collision.extended<-function(R,blade.profile,noBlades,birdLength,wingspan,flapping,birdSpeed,rpm,hub.height,xinc,y,py)          
{
   ymid<-y[-length(y)]+diff(y)/2
   ny<-length(ymid)
   risk.upwind<-rep(0,ny)
   risk.downwind<-rep(0,ny)
   for( i in 1:ny)
   {
      xmax <-sqrt(1 - ymid[i] ^2)
      n <-floor(xmax / xinc)
      x<-unique(c((0:n)*xinc,xmax))
      p.r<-sqrt(x^2+ymid[i]^2)
      r<-p.r*R
      phi<-atan(x/ymid[i])
      phi[is.nan(phi)]<-pi/2
      c<-interpolate(from.x=blade.profile$pR,from.y=blade.profile$c,to.x=p.r)
      pitch<-interpolate(from.x=blade.profile$pR,from.y=blade.profile$pitch,to.x=p.r)
 
      upwind.collision.risk<-pcoll(r,c,phi,pitch,noBlades,birdLength,wingspan,flapping,birdSpeed,rpm,upwind=TRUE)
      downwind.collision.risk<-pcoll(r,c, phi,pitch,noBlades,birdLength,wingspan,flapping,birdSpeed,rpm,upwind=FALSE)

      risk.upwind[i]<-trap.rule(x,upwind.collision.risk$p.collide)/xmax
      risk.downwind[i]<-trap.rule(x,downwind.collision.risk$p.collide)/xmax
   }
            
   p.collision.upwind<-sum(py*risk.upwind)/sum(py)
   p.collision.downwind<-sum(py*risk.downwind)/sum(py)
   p.collision<-mean(c(p.collision.upwind,p.collision.downwind))
   p.collision
}
#
# This function returns the cumulative distribution function for a variety of theoretical flight height distributions (as fitted to the site-specific data in 
# The Hywind project) and also the SOSS generic flight height distributions. Provides the theoretical framework which allows the same software to be used
# to estimate collision risk mortality across all flight height distributions.  
#
#
get.cdf<-function(x,best.fit,distribution)
{                            
 if(distribution=="exp")
 {
  cdf<-pexp(x,rate=best.fit$rate)
 } 
 else if(distribution=="gamma")
 {
  cdf<-pgamma(x,shape=best.fit$shape,rate=best.fit$rate)
 } 
 else if(distribution=="lnorm")
 {
    cdf<-plnorm(x,meanlog=best.fit$meanlog,sdlog=best.fit$sdlog)
 } 
 else if(distribution=="pareto")
 {
    cdf<-ppareto(x,shape=best.fit$shape,scale=best.fit$scale)
 } 
 else if(distribution=="genpareto")
 {
    cdf<-pgenpareto(x,shape1=best.fit$shape1,shape2=best.fit$shape2, scale=best.fit$scale)
 } 
 else if(distribution=="SOSS")
 {         
    cdf<-interpolate(from.x=best.fit$height,from.y=best.fit$cdf,to.x=x)
 }
 cdf
}
#
# This function calculates the probability of collision for the Basic model for both the stage 1 and stage 2 software
# 
calculate.p.collision<-function(R,blade.profile,noBlades,birdLength,wingspan,flapping,birdSpeed,rpm=NULL,n=20)
{
  p.r<-(0:n/n)
  r<-p.r*R
  c<-interpolate(from.x=blade.profile$pR,from.y=blade.profile$c,to.x=p.r)
  pitch<-interpolate(from.x=blade.profile$pR,from.y=blade.profile$pitch,to.x=p.r)

 upwind.collision.risk<-pcoll(r,c,phi=NULL,pitch,noBlades,birdLength,wingspan,flapping,birdSpeed,rpm,upwind=TRUE)
 p.upwind<-upwind.collision.risk$p.collide
 upwind.collide.length<-upwind.collision.risk$collide.length 
 alpha<-upwind.collision.risk$alpha
 downwind.collision.risk<-pcoll(r,c, phi=NULL,pitch,noBlades,birdLength,wingspan,flapping,birdSpeed,rpm,upwind=FALSE)
 p.downwind<-downwind.collision.risk$p.collide
 downwind.collide.length<-downwind.collision.risk$collide.length
 
 p.upwind.overall<-(2/R^2)*trap.rule(r,r*p.upwind)
 p.downwind.overall<-(2/R^2)*trap.rule(r,r*p.downwind)
 calculation<-data.frame(r,c,alpha,upwind.collide.length,p.upwind,downwind.collide.length,p.downwind)    
 p.overall<-mean(c(p.upwind.overall,p.downwind.overall))
 list(p.upwind.overall=p.upwind.overall,p.downwind.overall=p.downwind.overall,p.overall=p.overall,calculation=calculation)
}
#
# This function performs linear interpolation
#
interpolate<-function(from.x,from.y,to.x)
{
  n<-length(to.x)
  below<-rep(0,n)
  above<-rep(0,n)
  for(i in 1:n)
  {
    below[i]<-max(which(to.x[i]>=from.x))
    splurge<-which(to.x[i]<from.x)
    if(length(splurge)>0)
    {
      above[i]<-min(splurge)
    }
    else
    {
      above[i]<-below[i]
    }  
  }
  delta_y<-from.y[above]-from.y[below]
  delta_x<-from.x[above]-from.x[below]
  delta_x[above==below]<-1
  c<-from.y[below]+(((to.x-from.x[below])/delta_x)*delta_y)
  c  
  
}
#
# This function calculates the probability of collision for a bird flying through the rotor at a particular point in the rotor plane for the extended model
# and for a particular radial distance along the blade for the basic model
#  It is used by all the collision risk models (i.e. basic and extended, stage 1 and stage 2)
# 
pcoll<-function(r,c, phi=NULL,pitch,noBlades,birdLength,wingspan,flapping,birdSpeed,rpm,upwind)
{
  b<-noBlades
  L<-birdLength
  w<-wingspan
  v<-birdSpeed
  omega<-rpm*2*pi/60 #convert rpm to radians/sec
# assumes birdSpeed in metres per second
  alpha<-birdSpeed/(r*omega)
  beta<-L/w
  gamma<-pitch*pi/180

  if(flapping)
  {
    F<-1
  }
  else
  {                    
    if(is.null(phi))
    {
      F<-2/pi
    }
    else
    {
      F=cos(phi)
    }
  }    
  theta<-w*alpha*F 
  
  if(is.null(phi))
  {
    theta[alpha<beta]<-L
  }
  else
  {
    theta<-pmax(L,theta)
  }
  
#  theta[alpha<beta]<-L  
#  print(data.frame(abs(c*sin(gamma)+ alpha*c*cos(gamma)),theta,abs(c*sin(gamma)+ alpha*c*cos(gamma)) + theta))
if(upwind)
{
  multiplier<-1
}
  else  
  {
   multiplier<--1
  }  
 collide.length<-abs(multiplier*c*sin(gamma)+ alpha*c*cos(gamma)) + theta
 p.collide<-(b*omega/(2*pi*v))*collide.length
 p.collide[p.collide>1]<-1
 if(!is.null(phi))
 {
   initial.parameters<-data.frame(r,c,phi)
 }
 else
 {
  initial.parameters<-data.frame(r,c)
 }
# print(paste("A",length(initial.parameters[,1]),length(omega),length(alpha),length(beta),length(gamma),length(collide.length),length(p.collide)))
 cbind(initial.parameters,data.frame(omega,alpha,beta,gamma,collide.length,p.collide))   
} 
#
# This function calculates Q2r simply and directly from the flight height distribution, as the difference between the proportion of birds below the top of the turbine,
# and the proportion below the bottom of the turbine. Used by the stage 2 software. 
#
calculateQ2r<-function(turbine.bottom,turbine.top,fhd.distribution,best.fit)    
{
    Q2r<-diff(get.cdf(x=c(turbine.bottom,turbine.top),distribution=fhd.distribution,best.fit=best.fit))
    Q2r
}    
#
# This function calculates Q2r according to Band's approximation. Used by the stage 1 software
#
calculate.Q2r.band<-function(R,hub.height,yinc,flightHeightDistribution)
{
  y<-calculate.y(yinc)
  dy<-calculate.dy(y,R,hub.height,flightHeightDistribution)
  trap.rule(y,dy)
}
#
# This function calculates the flux integral as defined in Band's guidance. Used in the stage 1 software
# 
calculate.flux.integral<-function(R,hub.height,yinc,flightHeightDistribution)
{
  y<-calculate.y(yinc)
  dy<-calculate.dy(y,R,hub.height,flightHeightDistribution)
  chord.length<-sqrt(1 - y^2)
  trap.rule(y,dy*chord.length)*4/pi
}
#
# This function calculates the collision integral as defined in Band's guidance. Used in the stage 1 software
#
calculate.collision.integral<-function(R,blade.profile,noBlades,birdLength,wingspan,flapping,birdSpeed,rpm,flightHeightDistribution,hub.height,xinc,yinc)
{
  y<-calculate.y(yinc)
  dy<-calculate.dy(y,R,hub.height,flightHeightDistribution)

  ny<-length(y)
  risk.upwind<-rep(0,ny)
  risk.downwind<-rep(0,ny)
  for( i in 1:ny)
  {
    xmax <-sqrt(1 - y[i] ^2)
    n <-floor(xmax / xinc)
    x<-unique(c((0:n)*xinc,xmax))
    p.r<-sqrt(x^2+y[i]^2)
    r<-p.r*R
    phi<-atan(x/y[i])
    phi[is.nan(phi)]<-pi/2
    c<-interpolate(from.x=blade.profile$pR,from.y=blade.profile$c,to.x=p.r)
    pitch<-interpolate(from.x=blade.profile$pR,from.y=blade.profile$pitch,to.x=p.r)
 
    upwind.collision.risk<-pcoll(r,c,phi,pitch,noBlades,birdLength,wingspan,flapping,birdSpeed,rpm,upwind=TRUE)
    downwind.collision.risk<-pcoll(r,c, phi,pitch,noBlades,birdLength,wingspan,flapping,birdSpeed,rpm,upwind=FALSE)
    risk.upwind[i]<-trap.rule(x,upwind.collision.risk$p.collide)
    risk.downwind[i]<-trap.rule(x,downwind.collision.risk$p.collide)
   }
   dy.risk.upwind<-dy*risk.upwind
   dy.risk.downwind<-dy*risk.downwind
   collision.integral.upwind<-trap.rule(y,dy.risk.upwind)*4/pi
   collision.integral.downwind<-trap.rule(y,dy.risk.downwind)*4/pi
   collision.integral<-mean(c(collision.integral.upwind,collision.integral.downwind))
   calculation<-data.frame(y,dy,risk.upwind,risk.downwind,dy.risk.upwind,dy.risk.downwind)
  list(collision.integral.upwind=collision.integral.upwind,collision.integral.downwind=collision.integral.downwind,collision.integral=collision.integral,calculation=calculation)
}
#
# This function calculates dy, as defined in Band's guidance. Used in the stage 1 soffware
#
calculate.dy<-function(y,R,hub.height,flightHeightDistribution)
{
  height<-y*R+hub.height
  dy<-R*interpolate(from.x=flightHeightDistribution$height,from.y=flightHeightDistribution$p,to.x=height)
}


